/*********************************************************************
*               SEGGER MICROCONTROLLER GmbH & Co KG                  *
*       Solutions for real time microcontroller applications         *
**********************************************************************
*                                                                    *
*       (c) 1995 - 2012  SEGGER Microcontroller GmbH & Co KG         *
*                                                                    *
*       www.segger.com     Support: support@segger.com               *
*                                                                    *
**********************************************************************
*                                                                    *
*       embOS * Real time operating system for microcontrollers      *
*                                                                    *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product or a real-time            *
*       operating system for in-house use.                           *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       OS version: Internal                                         *
*                                                                    *
*       Current version number will be inserted here                 *
*       when shipment is built.                                      *
*                                                                    *
**********************************************************************

----------------------------------------------------------------------
File    : RTOSInit_RM46L852.c for TI RM46L852 CPU and TI RM46 HDK eval board
          
Purpose : Initializes and handles the hardware for embOS as far
          as required by embOS
          Feel free to modify this file acc. to your target system.
--------  END-OF-HEADER  ---------------------------------------------
*/

#include "RTOS.h"

/*********************************************************************
*
*       Configuration
*
**********************************************************************
*/

/*********************************************************************
*
*       Clock frequency settings
*/
#ifndef   OS_FSYS                   /* CPU Main clock frequency     */
  #define OS_FSYS 220000000uL       /* Assume 220MHz                */
#endif

#ifndef   OS_PCLK_TIMER             /* Peripheral clock for timer   */
  #define OS_PCLK_TIMER (OS_FSYS/4) /* May vary from CPU clock      */
#endif                              /* depending on CPU             */

#ifndef   OS_PCLK_UART              /* Peripheral clock for UART    */
  #define OS_PCLK_UART (OS_FSYS)    /* May vary from CPU clock      */
#endif                              /* depending on CPU             */

#ifndef   OS_TICK_FREQ
  #define OS_TICK_FREQ (1000)
#endif

/*********************************************************************
*
*       Configuration of communication to OSView
*/
#ifndef   OS_VIEW_ENABLE            // Global enable of communication
  #define OS_VIEW_ENABLE    (1)     // Default: on
#endif

#ifndef   OS_VIEW_USE_UART          // If set, UART will be used for communication
  #define OS_VIEW_USE_UART  (0)     // Default: 0 => Uart is not used
#endif                              // if OS_VIEW_ENABLE is on

/*********************************************************************
*
*       UART settings for embOSView
*       If you do not want (or can not due to hardware limitations)
*       to dedicate a uart to embOSView, please define it to be -1
*       Currently no uart is supported.
*/
#ifndef   OS_UART
  #define OS_UART (-1)
#endif

#ifndef   OS_BAUDRATE
  #define OS_BAUDRATE (38400)
#endif

#ifndef   ALLOW_NESTED_INTERRUPTS
  #define ALLOW_NESTED_INTERRUPTS (0)
#endif

/****** End of configurable options *********************************/

/*********************************************************************
*
*       DCC and UART settings for embOSView
*       Automatically generated from configuration settings
*/
#define OS_USE_DCC     (OS_VIEW_ENABLE && (OS_VIEW_USE_UART == 0))

#define OS_UART_USED   (OS_VIEW_ENABLE && (OS_VIEW_USE_UART != -1))

#if OS_UART_USED
  #define OS_COM_INIT() OS_COM_Init()
#else
  #define OS_COM_INIT()
#endif

#if (OS_USE_DCC)
  #include "JLINKDCC.h"
#endif

#define OS_TIMER_RELOAD (OS_PCLK_TIMER / OS_TICK_FREQ)

/*********************************************************************
*
*       Check configuration
*
**********************************************************************
*/

#ifndef   DEBUG     /* Should normally be defined as project option */
  #define DEBUG  (0)
#endif

/*********************************************************************
*
*       Local defines (sfrs used in RTOSInit.c)
*
**********************************************************************
*/
#define PLLCTL1       (*(volatile unsigned int*)(0xFFFFFF70))   // PLL Control 1 Register 
#define PLLCTL2       (*(volatile unsigned int*)(0xFFFFFF74))   // PLL Control 2 Register 
#define PLLCTL3       (*(volatile unsigned int*)(0xFFFFE100))   // PLL Control 3 Register 
#define CSDIS         (*(volatile unsigned int*)(0xFFFFFF30))   // Clock Source Disable Register
#define CLKCNTL       (*(volatile unsigned int*)(0xFFFFFFD0))   // Clock Control Register
#define CLK2CNTRL     (*(volatile unsigned int*)(0xFFFFE13C))   // Clock 2 Control Register
#define PENA_MASK     (1u << 8)
#define PSPWRDWNCLR0  (*(volatile unsigned int*)(0xFFFFE0A0))   // Peripheral Power-Down Clear Register 0
#define PSPWRDWNCLR1  (*(volatile unsigned int*)(0xFFFFE0A4))   // Peripheral Power-Down Clear Register 1
#define PSPWRDWNCLR2  (*(volatile unsigned int*)(0xFFFFE0A8))   // Peripheral Power-Down Clear Register 2
#define PSPWRDWNCLR3  (*(volatile unsigned int*)(0xFFFFE0AC))   // Peripheral Power-Down Clear Register 3
#define FRDCNTL       (*(volatile unsigned int*)(0xFFF87000))   // Flash Option Control Register 
#define FSM_WR_ENA    (*(volatile unsigned int*)(0xFFF87288))   // FSM Register Write Enable 
#define EEPROM_CONFIG (*(volatile unsigned int*)(0xFFF872B8))   // EEPROM Emulation Configuration Register
#define FBFALLBACK    (*(volatile unsigned int*)(0xFFF87040))   // Flash Bank Fallback Power Register
#define CDDIS         (*(volatile unsigned int*)(0xFFFFFF3C))   // Clock Domain Disable Register
#define CSVSTAT       (*(volatile unsigned int*)(0xFFFFFF54))   // Clock Source Valid Status Register
#define GHVSRC        (*(volatile unsigned int*)(0xFFFFFF48))   // GCLK, HCLK, VCLK, and VCLK2 Source Register
#define RCLKSRC       (*(volatile unsigned int*)(0xFFFFFF50))   // RTI Clock Source Register
#define VCLKASRC      (*(volatile unsigned int*)(0xFFFFFF4C))   // Peripheral Asynchronous Clock Source Register
#define VCLKACON1     (*(volatile unsigned int*)(0xFFFFE140))   // Peripheral Asynchronous Clock Configuration 1 Register
#define VCLKR_MASK    (1u << 16)
#define VCLK2R_MASK   (1u << 24)
#define VCLK3R_MASK   (1u << 0)
#define TRUE          (1u)
#define FALSE         (0u)

enum flashWPowerModes {
  SYS_SLEEP   = 0U, // Alias for flash bank power mode sleep
  SYS_STANDBY = 1U, // Alias for flash bank power mode standby
  SYS_ACTIVE  = 3U  // Alias for flash bank power mode active
};
enum systemClockSource {
  SYS_OSC           = 0,  /**< Alias for oscillator clock Source                */
  SYS_PLL1          = 1,  /**< Alias for Pll1 clock Source                      */
  SYS_EXTERNAL1     = 3,  /**< Alias for external clock Source                  */
  SYS_LPO_LOW       = 4,  /**< Alias for low power oscillator low clock Source  */
  SYS_LPO_HIGH      = 5,  /**< Alias for low power oscillator high clock Source */
  SYS_PLL2          = 6,  /**< Alias for Pll2 clock Source                      */
  SYS_EXTERNAL2     = 7,  /**< Alias for external 2 clock Source                */
  SYS_VCLK          = 9   /**< Alias for synchronous VCLK1 clock Source         */
};

#define VCLKACON1      (*(volatile unsigned int*)(0xFFFFE140))   // Peripheral Asynchronous Clock Configuration 1 Register
#define RTIGCTRL       (*(volatile unsigned int*)(0xFFFFFC00))   // RTI Global Control Register Section
#define RTITBCTRL      (*(volatile unsigned int*)(0xFFFFFC04))   // RTI Timebase Control Register Section
#define RTICAPCTRL     (*(volatile unsigned int*)(0xFFFFFC08))   // RTI Capture Control Register Section
#define RTICOMPCTRL    (*(volatile unsigned int*)(0xFFFFFC0C))   // RTI Compare Control Register Section
#define RTIFRC0        (*(volatile unsigned int*)(0xFFFFFC10))   // RTI Free Running Counter 0 Register Section
#define RTIUC0         (*(volatile unsigned int*)(0xFFFFFC14))   // RTI Up Counter 0 Register Section
#define RTICPUC0       (*(volatile unsigned int*)(0xFFFFFC18))   // RTI Compare Up Counter 0 Register Section
#define RTICAFRC0      (*(volatile unsigned int*)(0xFFFFFC20))   // RTI Capture Free Running Counter 0 Register Section
#define RTICAUC0       (*(volatile unsigned int*)(0xFFFFFC24))   // RTI Capture Up Counter 0 Register Section
#define RTICOMP0       (*(volatile unsigned int*)(0xFFFFFC50))   // RTI Compare 0 Register Section
#define RTIUDCP0       (*(volatile unsigned int*)(0xFFFFFC54))   // RTI Update Compare 0 Register Section
#define RTISETINTENA   (*(volatile unsigned int*)(0xFFFFFC80))   // RTI Set Interrupt Enable Register Section
#define RTIINTFLAG     (*(volatile unsigned int*)(0xFFFFFC88))   // RTI Interrupt Flag Register Section

#define RTI_INT_ID           (2u)
#define RTIINTFLAG_INT0_MASK (1u << 0) 
#define VIM_CHANNEL          (96u)
#define VIM_RAM_ADDR         (0xFFF82000u) // Fixed vector table start addr in RAM

#define FIRQPR0        (*(volatile unsigned int*)(0xFFFFFE10))   // FIQ/IRQ Program Control Register 0
#define FIRQPR1        (*(volatile unsigned int*)(0xFFFFFE14))   // FIQ/IRQ Program Control Register 1
#define FIRQPR2        (*(volatile unsigned int*)(0xFFFFFE18))   // FIQ/IRQ Program Control Register 2
#define REQENASET0     (*(volatile unsigned int*)(0xFFFFFE30))   // Interrupt Enable Set Register 0
#define REQENASET1     (*(volatile unsigned int*)(0xFFFFFE34))   // Interrupt Enable Set Register 1
#define REQENASET2     (*(volatile unsigned int*)(0xFFFFFE38))   // Interrupt Enable Set Register 2
#define REQENACLR0     (*(volatile unsigned int*)(0xFFFFFE40))   // Interrupt Enable CLR Register 0
#define REQENACLR1     (*(volatile unsigned int*)(0xFFFFFE44))   // Interrupt Enable CLR Register 1
#define REQENACLR2     (*(volatile unsigned int*)(0xFFFFFE48))   // Interrupt Enable CLR Register 2
#define IRQVECREG      (*(volatile unsigned int*)(0xFFFFFE70))   // IRQ Interrupt Vector Register

/*********************************************************************
*
*       Static data
*
**********************************************************************
*/

/*********************************************************************
*
*       Local functions
*
**********************************************************************
*/

/********************************************************************* 
* 
*       _PLL_Init()
* 
* Function description 
*   Initializes the PLL
*/ 
static void _PLL_Init(void) {
  PLLCTL1 = 0x00000000U 
          | 0x20000000U 
          | ((0x1F) << 24)      // PLLDIV 0x20 = 32
          | 0x00000000U 
          | ((6u - 1u) << 16)   // REFCLKDIV
          | ((165u - 1u) << 8); // PLLMUL

  PLLCTL2 = 0x00000000U
          | (255u << 22)
          | (7u << 12)
          | ((2u - 1u) << 9)    // ODPLL
          | 61u;

  PLLCTL3 = ((2u - 1u) << 29)   // ODPLL
          | ((0x1Fu) << 24)     // PLLDIV
          | ((6u - 1u) << 16) 
          | ((165u - 1u) << 8);
    
  CSDIS = 0x00000000U 
        | 0x00000000U 
        | 0x00000008U 
        | 0x00000080U 
        | 0x00000000U 
        | 0x00000000U 
        | 0x00000000U;
  //
  // Disable Peripherals before peripheral powerup
  //
  CLKCNTL &= ~PENA_MASK;
  //
  // Release peripherals from reset and enable clocks to all peripherals 
  // Power-up all peripharals
  //
  PSPWRDWNCLR0 = 0xFFFFFFFFu;
  PSPWRDWNCLR1 = 0xFFFFFFFFu;
  PSPWRDWNCLR2 = 0xFFFFFFFFu;
  PSPWRDWNCLR3 = 0xFFFFFFFFu;
  //
  // Enable Peripherals
  //
  CLKCNTL |= PENA_MASK;
}

/********************************************************************* 
* 
*       _Flash_Init()
* 
* Function description 
*   Setup flash
*/ 
static void _Flash_Init(void) {
  //
  // Setup flash read mode, address wait states and data wait states */
  //
  FRDCNTL = 0x00000000U
          | (3U << 8U)
          | (1U << 4U)
          |  1U;
  //
  // Setup flash access wait states for bank 7
  //
  FSM_WR_ENA    = 0x5;
  EEPROM_CONFIG = 0x00030002;
  FSM_WR_ENA    = 0xA;          // Disable write access to flash state machine registers
  //
  // Setup flash bank power modes
  //
  FBFALLBACK = 0x00000000
             | (SYS_ACTIVE << 14U)
             | (SYS_SLEEP << 12U)
             | (SYS_SLEEP << 10U)
             | (SYS_SLEEP << 8U)
             | (SYS_SLEEP << 6U)
             | (SYS_SLEEP << 4U)
             | (SYS_ACTIVE << 2U)
             |  SYS_ACTIVE;
}

/********************************************************************* 
* 
*       _MapClocks()
* 
* Function description 
*   Initialize Clock Tree 
*/ 
static void _MapClocks(void) {
  //
  // Initialize Clock Tree: 
  // Disable / Enable clock domain
  //  
  CDDIS = (FALSE << 4 )  /* AVCLK 1 OFF */
         |(TRUE << 5 )   /* AVCLK 2 OFF */
         |(FALSE << 8 )  /* VCLK3 OFF   */
         |(FALSE << 10)  /* AVCLK 3 OFF */
         |(FALSE << 11); /* AVCLK 4 OFF */
  //
  // Wait for until clocks are locked
  //
  while ((CSVSTAT & ((CSDIS ^ 0xFF) & 0xFF)) != ((CSDIS ^ 0xFF) & 0xFF)) {
  }
  //
  // Now the PLLs are locked and the PLL outputs can be sped up
  // The R-divider was programmed to be 0xF. Now this divider is changed to programmed value
  //
  PLLCTL1 = (PLLCTL1 & 0xE0FFFFFF)|((1u - 1u)<< 24u);
  PLLCTL3 = (PLLCTL3 & 0xE0FFFFFF)|((1u - 1u)<< 24u);

  /** - Map device clock domains to desired sources and configure top-level dividers */
  /** - All clock domains are working off the default clock sources until now */
  /** - The below assignments can be easily modified using the HALCoGen GUI */

  /** - Setup GCLK, HCLK and VCLK clock source for normal operation, power down mode and after wakeup */
  GHVSRC = (SYS_PLL1 << 24U)
         | (SYS_PLL1 << 16U)
         |  SYS_PLL1;
  //
  // Setup synchronous peripheral clock dividers for VCLK1, VCLK2, VCLK3 */
  //
  CLKCNTL   |= VCLKR_MASK | VCLK2R_MASK;
  CLK2CNTRL |= VCLK3R_MASK;
  //
  // Setup RTICLK1 and RTICLK2 clocks
  //
  RCLKSRC = (1U << 24U)
          | (SYS_VCLK << 16U)
          | (1U << 8U)
          |  SYS_VCLK;

  // Setup asynchronous peripheral clock sources for AVCLK1 and AVCLK2
  VCLKASRC  = (SYS_VCLK << 8U) | SYS_VCLK;
  VCLKACON1 = (1U << 24)
            | 1 << 20U
            | (SYS_VCLK << 16)
            | (1U << 8)
            | 1 << 4U
            | SYS_VCLK; 
}

/*********************************************************************
*
*       _DefaultInterrupt_Handler()
*
* Function description
*   Default interrupt handler
*/
static void _DefaultInterrupt_Handler(void) {
  //
  // Just enter an endless loop so that user can set a breakpoint here
  //
  while (1) {
  }
}

/********************************************************************* 
* 
*       _VIM_Init()
* 
* Function description 
*   Initializes the Vector interrupt manager
*/ 
static void _VIM_Init(void) {
  OS_ISR_HANDLER** papISR;
  OS_U32 i;
  //
  // Map all interrupts to IRQ
  //
  FIRQPR0 = 0u;
  FIRQPR1 = 0u;
  FIRQPR2 = 0u;
  //
  // Fill vector table with default interrupt handler 
  //
  papISR = (OS_ISR_HANDLER**)VIM_RAM_ADDR;
  for (i = 0 ; i < VIM_CHANNEL ; i++) {
    *(papISR + i) = &_DefaultInterrupt_Handler;
  }
}

/*********************************************************************
*
*       _SysTick_Handler()
*
* Function description
*   Systick handler
*/
static void _SysTick_Handler(void) {
  RTIINTFLAG = RTIINTFLAG_INT0_MASK; // Clear the pending interrupts flag
  OS_TICK_Handle();
  #if OS_USE_DCC
    DCC_Process();
  #endif
}

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_InitHW()
*
*       Initialize the hardware (timer) required for the OS to run.
*       May be modified, if an other timer should be used
*/
void OS_InitHW(void) {
  OS_IncDI();
  //
  // Initialize Vector interrupt manager
  //
  _VIM_Init();
  //
  // RTI Initialization
  //
  RCLKSRC      = 0x08;             // Clock source 8 i.e VClk, div the clk source 0 by 1
  RTIGCTRL     = 0x00000000;       // Disable RTIUC0 and RTIUC1
  RTIUC0       = 0x00000000;       // Initialize up Counter
  RTIFRC0      = 0x00000000;       // Initialize Free Running Counter
  RTICPUC0     = 0x00000001;       // Set Compare Up Counter value to Prescale RTICLK.
  RTICOMP0     = OS_TIMER_RELOAD;  // Compare value
  RTIUDCP0     = OS_TIMER_RELOAD;  // This value is added to generate new compare value since the timer register continous counting
  RTIINTFLAG   = 0x0000000F;       // Clearing the pending Interrupts Flags
  RTICOMPCTRL  = 0x00000000;       // Initializing the Compare Counter register
  RTISETINTENA = 0x00000001;       // Enabling compare interrupt 0
  RTIGCTRL     = 0x00000001;       // Enable the RTIUC0
  //
	// Set interrupt handler and enable RTI interrupt
	//
	OS_ARM_InstallISRHandler(RTI_INT_ID, _SysTick_Handler);
	OS_ARM_EnableISR(RTI_INT_ID);
  //
  // Initialize the optional UART for OS viewer
  //
  OS_COM_INIT();
  OS_DecRI();
}

/*********************************************************************
*
*       OS_Idle()
*
*       Please note:
*       This is basically the "core" of the idle loop.
*       This core loop can be changed, but:
*       The idle loop does not have a stack of its own, therefore no
*       functionality should be implemented that relies on the stack
*       to be preserved. However, a simple program loop can be programmed
*       (like toggeling an output or incrementing a counter)
*/
void OS_Idle(void) {    // Idle loop: No task is ready to execute
  while (1) {           // Nothing to do ... wait for interrupt
    #if (DEBUG == 0)
      #ifdef __ICCARM__  // IAR
        __asm(" wfi");
      #endif
      #ifdef __CC_ARM    // KEIL
        __wfi();
      #endif
      #ifdef __GNUC__    // GCC
        __asm(" wfi");
      #endif
    #endif
  }
}

/*********************************************************************
*
*       OS_GetTime_Cycles()
*
*       This routine is required for task-info via OSView or high
*       resolution time measurement functions.
*       It returns the system time in timer clock cycles.
*/
OS_U32 OS_GetTime_Cycles(void) {
  OS_U32 Time;
  OS_I32 Cnt;

  Time = OS_GetTime32();
	Cnt = OS_TIMER_RELOAD + RTIFRC0 - RTICOMP0;
  //
  // Check if timer interrupt pending ...
  //
  if (RTIINTFLAG & RTIINTFLAG_INT0_MASK) { // Interrupt pending, re-read timer and adjust result
    Cnt = OS_TIMER_RELOAD + RTIFRC0 - RTICOMP0; 
    Time++;
  }
  return (OS_TIMER_RELOAD * Time) + Cnt;
}

/*********************************************************************
*
*       OS_ConvertCycles2us()
*
*       Convert Cycles into micro seconds.
*
*       If your clock frequency is not a multiple of 1 MHz,
*       you may have to modify this routine in order to get proper
*       diagnostics.
*
*       This routine is required for profiling or high resolution time
*       measurement only. It does not affect operation of the OS.
*/
OS_U32 OS_ConvertCycles2us(OS_U32 Cycles) {
  return Cycles/(OS_PCLK_TIMER/1000000);
}

/*********************************************************************
*
*       Optional communication with embOSView
*
**********************************************************************
*/
#if OS_USE_DCC

/*********************************************************************
*
*       _DCC_OnRx()
*/
static void _DCC_OnRx(unsigned Channel, OS_U8 Data) {
  (void) Channel;   // Not supported by this version, suppress warning
  OS_OnRx(Data);
}

/*********************************************************************
*
*       _DCC_OnTx()
*/
static void _DCC_OnTx(unsigned Channel) {
  (void) Channel;   // Not supported by this version, suppress warning
  OS_OnTx();
}

/*********************************************************************
*
*       OS_COM_Init()
*       Initialize DCC for OSView
*/
void OS_COM_Init(void) {
  JLINKDCC_SetpfOnRx(1, _DCC_OnRx);
  JLINKDCC_SetpfOnTx(1, _DCC_OnTx);
}

/*********************************************************************
*
*       OS_COM_Send1()
*
*       Send 1 character via DCC for embOSView communication
*       Never call this function directly from your application
*/
void OS_COM_Send1(OS_U8 c) {
  JLINKDCC_SendCharOnChannelNB(1, c);
}

#else  /* UART for communication not used, define dummy functions */

void OS_COM_Send1(OS_U8 c) {
  OS_USEPARA(c);           /* Avoid compiler warning */
  OS_COM_ClearTxActive();  /* Let the OS know that Tx is not busy */
}
#endif

/*********************************************************************
*
*       OS interrupt handler and ISR specific functions
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_irq_handler()
*
*       Detect reason for IRQ and call correspondig service routine.
*       OS_irq_handler is called from IRQ_Handler function
*       found in RTOS.s
*/
void OS_irq_handler(void) {
  OS_ISR_HANDLER* pISR;
  pISR = (OS_ISR_HANDLER*) IRQVECREG;  // Read interrupt vector 
#if ALLOW_NESTED_INTERRUPTS
  OS_EnterNestableInterrupt();         // Now interrupts may be reenabled. If nesting should be allowed
#else
  OS_EnterInterrupt();                 // Inform OS that interrupt handler is running
#endif
  pISR();                              // Call interrupt service routine
#if ALLOW_NESTED_INTERRUPTS
  OS_LeaveNestableInterrupt();         // Leave nestable interrupt, perform task switch if required
#else
  OS_LeaveInterrupt();                 // Leave interrupt, perform task switch if required
#endif
}

/*********************************************************************
*
*       OS_ARM_InstallISRHandler()
*/
OS_ISR_HANDLER* OS_ARM_InstallISRHandler (int ISRIndex, OS_ISR_HANDLER* pISRHandler) {
  OS_ISR_HANDLER*  pOldHandler;
  OS_ISR_HANDLER** papISR;

#if OS_DEBUG
  if ((unsigned)ISRIndex >= VIM_CHANNEL) {
    OS_Error(OS_ERR_ISR_INDEX);
    return NULL;
  }
#endif
  OS_DI();
  papISR = (OS_ISR_HANDLER**)VIM_RAM_ADDR;
  pOldHandler          = *(papISR + (ISRIndex + 1u));
  *(papISR + (ISRIndex + 1)) = pISRHandler;
  OS_RestoreI();
  return pOldHandler;
}

/*********************************************************************
*
*       OS_ARM_EnableISR()
*/
void OS_ARM_EnableISR(int ISRIndex) {
#if OS_DEBUG
  if ((unsigned)ISRIndex >= VIM_CHANNEL) {
    OS_Error(OS_ERR_ISR_INDEX);
    return;
  }
#endif

  OS_DI();
  if (ISRIndex < 32) {
    REQENASET0 = (1u << ISRIndex);
  } else {
    if (ISRIndex < 64) {
      REQENASET1 = (1u << ISRIndex);
    } else {
      REQENASET2 = (1u << ISRIndex);
    }
  }  
  OS_RestoreI();
}

/*********************************************************************
*
*       OS_ARM_DisableISR()
*/
void OS_ARM_DisableISR(int ISRIndex) {
#if OS_DEBUG
  if ((unsigned)ISRIndex >= VIM_CHANNEL) {
    OS_Error(OS_ERR_ISR_INDEX);
    return;
  }
#endif

  OS_DI();
  if (ISRIndex < 32) {
    REQENACLR0 = (1u << ISRIndex);
  } else {
    if (ISRIndex < 64) {
      REQENACLR1 = (1u << ISRIndex);
    } else {
      REQENACLR2 = (1u << ISRIndex);
    }
  }  
  OS_RestoreI();
}

/*********************************************************************
*
*       __low_level_init()
*
*       Initialize clock generation and pll
*
*       Has to be modified, if another CPU clock frequency should be
*       used. This function is called during startup and
*       has to return 1 to perform segment initialization.
*
*       Because variables are not initialized before __low_level_init() is called,
*       no access or call which relies on any variable can be used here.
*/
#ifdef __cplusplus
extern "C" {
#endif
int __low_level_init(void);    // Avoid "no prototype" warning
#ifdef __cplusplus
  }
#endif
int __low_level_init(void) {
  //
  // Initialize PLL
  //
  _PLL_Init();
  //
  // Set Flash wait states
  _Flash_Init();
  //
  // Initialize Clock Tree
  //
  _MapClocks();
  
  return 1;               // Always return 1 to enable segment initilization !
}

/****** End Of File *************************************************/
